#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "ds00_lib.h"

/* bitmask for options[0] */
#define BOM0_CONN_USB_SIL    0x01
#define BOM0_CONN_USB_MICRO  0x02

#define BOM0_CONN_BDM        0x04
#define BOM0_CONN_LCD_SIL    0x08

#define BOM0_CONN_LCD_DIL1   0x10
#define BOM0_CONN_LCD_DIL2   0x20
#define BOM0_STUFF_DIL_PWR1  0x40 
#define BOM0_STUFF_DIL_PWR2  0x80
// E2 is assumed to be stuffed if DIL_PWR2 is set

#define BOM1_STUFF_BL        0x01
#define BOM1_STUFF_CONTR     0x02
#define BOM1_STUFF_BDM_RST   0x04
#define BOM1_STUFF_MAX232    0x08
#define BOM1_STUFF_LED1      0x10
#define BOM1_STUFF_LED2      0x20
#define BOM1_STUFF_LED3      0x40

#define BOM2_CONN_EXP_MASK   0x0F
#define BOM2_CONN_EXP_NONE   0x00
#define BOM2_CONN_EXP_DIL14  0x01
#define BOM2_CONN_EXP_SIL7F  0x02  // front connector SPI + GPIO0
#define BOM2_CONN_EXP_SIL7B  0x03  // back connector GPIO/TTL serial/RS232
#define BOM2_CONN_EXP_RS232  0x04  // 3-pin RS232 (12V) 
#define BOM2_CONN_EXP_RS232D 0x05  // 5-pin RS232 (TTL + 12V)

#define BOM2_STUFF_RW        0x10 // kills GPIO1/Rx
#define BOM2_STUFF_CS_PTB1   0x20 // E2
#define BOM2_STUFF_CS_PTA5   0x40 // BL
#define BOM2_STUFF_CS_PTA0   0x80 // CONTR

/* display[0] contains the display size */
#define DISP_COL_MASK 0x3F
#define DISP_ROW_MASK 0xC0
#define DISP_USER     0xFF

/* bitmask for display[1] */
#define DISP_COLOR_NONE  0x0
#define DISP_COLOR_BLACK 0x1
#define DISP_COLOR_WHITE 0x2
#define DISP_COLOR_RED   0x3
#define DISP_COLOR_GREEN 0x4
#define DISP_COLOR_BLUE  0x5
#define DISP_COLOR_ANY   0xF



static ds00_device_handle *devh = NULL;

/*
static void hexprint(int a, unsigned char *p, int s)
{
  int i;
  printf("%02x : ",a);
  for(i=0;i<s;i++) printf("%02x ",p[i]);
  printf("\n");
}*/

#define YN(a) (a?"YES":"NO")

static void print_color(unsigned int c)
{
  switch(c&0x0f) {
  case DISP_COLOR_NONE:
    printf("N/A");
    break;
  case DISP_COLOR_BLACK:
    printf("black");
    break;
  case DISP_COLOR_WHITE:
    printf("white");
    break;
  case DISP_COLOR_RED:
    printf("red");
    break;
  case DISP_COLOR_GREEN:
    printf("green");
    break;
  case DISP_COLOR_BLUE:
    printf("blue");
    break;
  default:
    printf("unknown");
  }
}

static int print_config(unsigned short *caps)
{
//  int i,j;
  printf("USB connector internal: %s\n", YN(caps[0]&BOM0_CONN_USB_SIL));
  printf("micro USB connector: %s\n", YN(caps[0]&BOM0_CONN_USB_MICRO));
  printf("BDM connector: %s\n", YN(caps[0]&BOM0_CONN_BDM));
  //  printf("BDM connector: %s\n", YN(caps[0]&BOM0_CONN_BDM));
 
  printf("display size: ");
  if((caps[2]&0xff)==0xff) printf("user\n");
  else printf("%dx%d\n",caps[2]&DISP_COL_MASK,1<<((caps[2]&DISP_ROW_MASK)>>6));

  printf("display color: ");
  print_color(caps[2]>>12);
  printf("\n");

  printf("backlight color: "); 
  print_color(caps[2]>>8);
  printf("\n");

  return 0;
}

static int print_dev_info(void)
{
  int config, version;

  config=ds00_get_config(devh);
  version=ds00_get_fw_version(devh);

  printf("config string %04x\n",config);
  printf("Firmware version %x.%02x\n",version&0xff,version>>8);


  return 0;
}

int ds00_send(ds00_device_handle *devh, int cmd, int value);

int main(int argc, char *argv[])
{
  int r;
  
  r = ds00_usb_init(NULL);
  if (r < 0) {
    fprintf(stderr, "Failed to initialise libusb\n");
    exit(1);
  }
  //    libusb_set_debug(NULL,3);
  
  devh = ds00_open();
  if (devh==NULL) {
    fprintf(stderr, "Could not find/open DS00\n");
    goto out;
  }
  //  printf("Successfully found DS00\n");
  
  
  if(argc==2 && *argv[1]=='p') {
    unsigned short caps[3];
    caps[0]=ds00_get_caps(devh,0);
    caps[1]=ds00_get_caps(devh,2);
    caps[2]=ds00_get_caps(devh,4);

    printf("caps: ");
    printf("%04x %04x %04x\n", caps[0], caps[1], caps[2] );
    print_config(caps);

  } else if(argc==3 && *argv[1]=='b') {
    ds00_set_backlight(devh,atoi(argv[2]));
  } else if(argc==3 && *argv[1]=='t') {
    ds00_set_contrast(devh,atoi(argv[2]));
  } else if(argc==2 && *argv[1]=='i') {
    print_dev_info();
  } else if(argc==2 && *argv[1]=='S') {
    ds00_cfg_save(devh);
  } else if(argc==2 && *argv[1]=='W') {
    int config;
    config=ds00_get_config(devh);
    ds00_set_config(devh,config|0x80);
    ds00_cfg_save(devh);
    ds00_reset(devh,0);
  } else if(argc==2 && *argv[1]=='L') {
    int config;
    config=ds00_get_config(devh);
    ds00_set_config(devh,config&0xffffff7f);
    ds00_cfg_save(devh);
    ds00_reset(devh,0);
  } else if(argc==2 && *argv[1]=='k') {
    printf("key status: %04x\n",ds00_get_keys(devh));
    //	  print_dev_info();
  } else if(argc==3 && *argv[1]=='f') {
    ds00_set_config(devh,atoi(argv[2]));
  } else if(argc==3 && *argv[1]=='g') {
    ds00_set_gpio(devh,atoi(argv[2]));
  } else if(argc==2 && *argv[1]=='x') {
    ds00_reset(devh,0);
  } else if(argc==2 && *argv[1]=='r') {
    ds00_reset(devh,3);
  } else if(argc==2 && *argv[1]=='u') {
    ds00_reset(devh,0xC3);
  } else if(argc>=3 && *argv[1]=='d') {
    switch(argc) {
    case 3:
      ds00_send(devh,0x58,atoi(argv[2]));
      break;
    case 4:
      ds00_send(devh,0x59,atoi(argv[2])|(atoi(argv[3])<<8));
      break;
    case 5:
      //      ds00_send(devh,0x5a,atoi(argv[2])|(atoi(argv[3])<<8),atoi(argv[4]));
      break;
    case 6:
      //      ds00_send(devh,0x5b,atoi(argv[2])|(atoi(argv[3])<<8),atoi(argv[4])|(atoi(argv[5])<<8));
      break;
    }
  } else if(argc==3 && *argv[1]=='c') {
    ds00_send(devh,0x38,atoi(argv[2]));
    /*	} else if(argc==3 && *argv[1]=='t') {
	ds00_send_cmd1(devh,DS00_CMD_SET_CONT,atoi(argv[2]));
	} else if(argc==3 && *argv[1]=='b') {
	ds00_send_cmd1(devh,DS00_CMD_SET_BL,atoi(argv[2]));
	} else if(argc==4 && *argv[1]=='w') {
	ds00_reg_write(devh,atoi(argv[2]),atoi(argv[3]));*/
  } else {
    printf("usage:\n");
    printf("ds00_cli i     print device info\n");
    //    printf("ds00_cli v     print firmware version\n");
    printf("ds00_cli k     print button status\n");
    
    
    printf("ds00_cli x     reboot MCU\n");
    printf("ds00_cli u     upgrade firmware\n");
    //	  printf("ds00_cli k     kill firmware. will drop in bootloader next reset\n");
    
    //	  printf("ds00_cli c     clear display\n");
    //	  printf("ds00_cli h     home\n");
    printf("ds00_cli d xx  send one data byte\n");
    printf("ds00_cli c xx  send one cmd byte\n");
    printf("ds00_cli t xx  set contrast\n");
    printf("ds00_cli b xx  set backlight\n");
    printf("ds00_cli g xx  set gpio\n");
    printf("ds00_cli f xx  set config\n");
    printf("ds00_cli p     get caps\n");
    printf("ds00_cli S     save current settings to nvram\n");
    printf("ds00_cli W     switch to HID mode and save to nvram\n");
    printf("ds00_cli L     switch to lcd2usb mode and save to vram\n");
    
    /*	  printf("ds00_cli l     load config from flash\n");
	  printf("ds00_cli s     save config to flash\n");
	  printf("ds00_cli w a d   write d into register a\n");
	  printf("ds00_cli c x dd dd dd dd dd dd dd dd    configure fan x\n"); */

	}

//	test_interrupt_transfer();
out:
  ds00_close(devh);
  ds00_usb_exit(NULL);
  return 0;
}
