#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef USE_OLD_LIBUSB
#include <libusb-1.0/libusb.h>
#define USB_DEVH libusb_device_handle
#define USB_INIT libusb_init
#define USB_CTRL_TRANSFER(p1,p2,p3,p4,p5,p6,p7,p8) libusb_control_transfer(p1,p2,p3,p4,p5,p6,p7,p8)
#define USB_SET_CONFIG libusb_set_configuration
#define USB_CLAIM libusb_claim_interface
#define USB_RELEASE libusb_release_interface
#define USB_CLOSE libusb_close
#else
#include <usb.h>
#define USB_DEVH usb_dev_handle
#define USB_INIT usb_init
#define USB_CTRL_TRANSFER(p1,p2,p3,p4,p5,p6,p7,p8) usb_control_msg(p1,p2,p3,p4,p5,(char *)p6,p7,p8)
#define USB_SET_CONFIG usb_set_configuration
#define USB_CLAIM usb_claim_interface
#define USB_RELEASE usb_release_interface
#define USB_CLOSE usb_close
#endif

//#include "ds00_usb_interface.h"
#include "ds00_lib.h"

#define VERSION "0.1.0"

/* vendor and product id */
#define LCD2USB_VID  0x0403
#define LCD2USB_PID  0xc630
#define HID_VID      LCD2USB_VID
#define HID_PID      0xfc63

/* target is a bit map for CMD/DATA */
#define LCD_CTRL_0         (1<<3)
#define LCD_CTRL_1         (1<<4)
#define LCD_BOTH           (LCD_CTRL_0 | LCD_CTRL_1)

#define LCD_ECHO           (0<<5)
#define LCD_CMD            (1<<5)
#define LCD_DATA           (2<<5)
#define LCD_SET            (3<<5)
#define LCD_GET            (4<<5)
#define LCD_SPI_CMD        (6<<5)
#define LCD_SYS_CMD        (7<<5)

/* LCD_SET target is value to set */
#define LCD_SET_CONTRAST   (LCD_SET | (0<<3))
#define LCD_SET_BRIGHTNESS (LCD_SET | (1<<3))
#define LCD_SET_RESERVED0  (LCD_SET | (2<<3))
#define LCD_SET_RESERVED1  (LCD_SET | (3<<3))

/* LCD_GET target is value to get */
#define LCD_GET_FWVER      (LCD_GET | (0<<3))
#define LCD_GET_KEYS       (LCD_GET | (1<<3))
#define LCD_GET_CTRL       (LCD_GET | (2<<3))
#define LCD_GET_RESERVED1  (LCD_GET | (3<<3))

/* SPI operations */
#define LCD_SPI_IO       (LCD_SPI_CMD|<<3)

/* LCD_SYS_CMD target */
#define LCD_GET_CFG  (LCD_SYS_CMD|0<<3)
#define LCD_SET_CFG  (LCD_SYS_CMD|0<<3|2)
//#define LCD_DSCMD_GET_GPIO  (LCD_DSCMD|0<<3)
#define LCD_SET_GPIO     (LCD_SYS_CMD|1<<3|2)

#define LCD_GET_PCB_FEATURES (LCD_SYS_CMD|1<<3)
#define LCD_SAVE_CFG         (LCD_SYS_CMD|2<<3)
#define LCD_RESET       (LCD_SYS_CMD|3<<3)

#define HID_GET_REPORT                0x01
#define HID_GET_IDLE                  0x02
#define HID_GET_PROTOCOL              0x03
#define HID_SET_REPORT                0x09
#define HID_SET_IDLE                  0x0A
#define HID_SET_PROTOCOL              0x0B
#define HID_REPORT_TYPE_INPUT         0x01
#define HID_REPORT_TYPE_OUTPUT        0x02
#define HID_REPORT_TYPE_FEATURE       0x03

//#define LCD_CMD_SET_SPI  (LCD_DSCMD|2<<3|2)
//#define LCD_CMD_GET_SERIAL  (LCD_DSCMD|3<<3)
//#define LCD_CMD_SET_SERIAL  (LCD_DSCMD|3<<3|2)
#ifndef USE_OLD_LIBUSB
#define CTRL_IN	   (LIBUSB_ENDPOINT_IN|LIBUSB_REQUEST_TYPE_VENDOR|LIBUSB_RECIPIENT_DEVICE)
#define CTRL_OUT   (LIBUSB_ENDPOINT_OUT|LIBUSB_REQUEST_TYPE_VENDOR|LIBUSB_RECIPIENT_DEVICE)
#define HID_IN	   (LIBUSB_ENDPOINT_IN|LIBUSB_REQUEST_TYPE_CLASS|LIBUSB_RECIPIENT_DEVICE)
#define HID_OUT	   (LIBUSB_ENDPOINT_OUT|LIBUSB_REQUEST_TYPE_CLASS|LIBUSB_RECIPIENT_DEVICE)
#else
#define CTRL_IN	   (USB_ENDPOINT_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE)
#define CTRL_OUT   (USB_ENDPOINT_OUT|USB_YPE_VENDOR|USB_RECIP_DEVICE)
#define HID_IN	   (USB_ENDPOINT_IN|USB_TYPE_CLASS|USB_RECIP_DEVICE)
#define HID_OUT	   (USB_ENDPOINT_OUT|USB_TYPE_CLASS|USB_RECIP_DEVICE)
#endif

#define CMD_PACKET_SIZE 8

//#define PACKET_SIZE RFEAT_FAN_CFG_SIZE

#define ENDPOINT_INT_IN  0x81  /* endpoint 0x81 address for IN */
#define ENDPOINT_INT_OUT 0x01  /* endpoint 1 address for OUT */

const static int TIMEOUT=5000; /* timeout in ms */

int ds00_send_control(USB_DEVH *devh, int cmd, int value);
int ds00_recv_control(USB_DEVH *devh, int cmd, int idx);
int ds00_send_report(USB_DEVH *devh, int cmd, int value);
int ds00_recv_report(USB_DEVH *devh);
int ds00_find_device(ds00_device_handle *ds00);


int ds00_usb_init(void **ctx)
{
#ifndef USE_OLD_LIBUSB
  return libusb_init((libusb_context **)ctx);
#else
  if(ctx) *ctx=NULL;
  usb_init();
  return 0;
#endif
}

void ds00_usb_exit(void *ctx)
{
#ifndef USE_OLD_LIBUSB
  libusb_exit((libusb_context *)ctx);
#else
  //  usb_exit();
#endif
}

int ds00_find_device(ds00_device_handle *ds00)
{
  USB_DEVH *devh;

#ifndef USE_OLD_LIBUSB
  devh = libusb_open_device_with_vid_pid(NULL, LCD2USB_VID, LCD2USB_PID);
  if(devh!=NULL) {
    ds00->pid=LCD2USB_PID;
  } else {
    devh = libusb_open_device_with_vid_pid(NULL, HID_VID, HID_PID);
    if(devh!=NULL) {
      ds00->pid=HID_PID;
      //      ds00->pid=LCD2USB_PID;
    } else {
      return -1;
    }
  }
#else
  struct usb_bus      *bus;
  struct usb_device   *dev;

  devh=NULL;

  usb_find_busses();
  usb_find_devices();

  printf("successfully enumerated buses\n");
  for(bus = usb_get_busses(); bus; bus = bus->next) {
    for(dev = bus->devices; dev; dev = dev->next) {
    printf("considering vid:%04x pid:%04x\n",dev->descriptor.idVendor,
	   dev->descriptor.idProduct);
      if((dev->descriptor.idVendor == LCD2USB_VID) &&
	 (dev->descriptor.idProduct == LCD2USB_PID)) {
	/* open device */
	devh = usb_open(dev);
	ds00->pid=LCD2USB_PID;
	break;
      }
      if((dev->descriptor.idVendor == HID_VID) &&
	 (dev->descriptor.idProduct == HID_PID)) {
	/* open device */
	printf("found hid device\n");
	devh = usb_open(dev);
	ds00->pid=HID_PID;
	break;
      }
    }
  }
  if(devh==NULL) return -1;
#endif

  ds00->dev_handle=devh;
  return 0;
}

ds00_device_handle *ds00_open(void)
{
  //  USB_DEVH *devh;
  ds00_device_handle *ds00;
  int r;

  ds00=malloc(sizeof(ds00_device_handle));
  if(ds00==NULL) return NULL;

  r=ds00_find_device(ds00);

  if(r<0) {
    free(ds00);
    return NULL;
  }

#ifndef WIN32
  r= libusb_kernel_driver_active(ds00->dev_handle, 0);
  if(r) {
    r= libusb_detach_kernel_driver(ds00->dev_handle, 0);
    if (r < 0) {
      fprintf(stderr, "cannot detach kernel driver %d\n", r);
      goto out;
    }
  }
#endif

  r = USB_SET_CONFIG(ds00->dev_handle, 1);
  if (r < 0) {
    fprintf(stderr, "libusb_set_configuration error %d\n", r);
    goto out;
  }
      printf("Successfully set usb configuration 1\n");
  r = USB_CLAIM(ds00->dev_handle, 0);
  if (r < 0) {
    fprintf(stderr, "libusb_claim_interface error %d\n", r);
    goto out;
  }
     printf("Successfully claimed interface\n");

  return ds00;
 out:
  USB_CLOSE(ds00->dev_handle);
  free(ds00);
  return NULL;
}


int ds00_close(ds00_device_handle *devh)
{
  if(devh) {
    USB_RELEASE(devh->dev_handle, 0);
    USB_CLOSE(devh->dev_handle);
    free(devh);
    return 0;
  }
  return -1;
}


/*************************************************************/
/*                                */
/*************************************************************/
int ds00_send(ds00_device_handle *devh, int cmd, int value)
{
  if(devh->pid==HID_PID) {
//     printf("sending report\n");
    return ds00_send_report(devh->dev_handle, cmd, value);
  } else {
//     printf("sending control\n");
    return ds00_send_control(devh->dev_handle, cmd, value);
  }
}

int ds00_recv(ds00_device_handle *devh, int cmd, int idx)
{
  if(devh->pid==HID_PID) {
//    printf("in ds00_recv\n");
    ds00_send_report(devh->dev_handle, cmd, idx);
    return ds00_recv_report(devh->dev_handle);
  } else {
    return ds00_recv_control(devh->dev_handle, cmd, idx);
  }
}


/*************************************************************/
/* Control Transfer Interface                                */
/*************************************************************/
int ds00_send_control(USB_DEVH *devh, int cmd, int value)
{
  return USB_CTRL_TRANSFER(devh, CTRL_IN, cmd, value, 0,
			   NULL, 0, TIMEOUT);
}

int ds00_recv_control(USB_DEVH *devh, int cmd, int idx)
{
  unsigned char buf[2];
  int r;

  r = USB_CTRL_TRANSFER(devh, CTRL_IN, cmd, idx, 0,
			(unsigned char *)buf, sizeof(buf), TIMEOUT);

  if(r<0) return -1;

  return buf[0]+256*buf[1];
}

/*************************************************************/
/* HID Interface                                             */
/*************************************************************/
int ds00_recv_report(USB_DEVH *devh)
{
  unsigned char buf[CMD_PACKET_SIZE];
  int r;

  r = USB_CTRL_TRANSFER(devh, HID_IN, HID_GET_REPORT,
			(HID_REPORT_TYPE_FEATURE<<8),
			0, buf, CMD_PACKET_SIZE, TIMEOUT);

  if (r < 0) {
    fprintf(stderr, "Control IN error %d\n", r);
    return r;
  }

  r=buf[0]+(buf[1]<<8);
  return r;
}

int ds00_send_report(USB_DEVH *devh, int cmd, int value)
{
  unsigned char buf[CMD_PACKET_SIZE];
  int r;

  buf[0]=cmd&0xff;
  buf[1]=value&0xff;
  buf[2]=(value>>8)&0xff;

  r = USB_CTRL_TRANSFER(devh, HID_OUT, HID_SET_REPORT,
			(HID_REPORT_TYPE_FEATURE<<8),
			0, buf, CMD_PACKET_SIZE, TIMEOUT);

  if (r < 0) {
    fprintf(stderr, "Control Out error %d\n", r);
    return r;
  }
  return 0;
}


/*************************************************************/
/* User calls                                                */
/*************************************************************/
int ds00_set_contrast(ds00_device_handle *devh, int value)
{
  return ds00_send(devh,LCD_SET_CONTRAST,value);
}

int ds00_set_backlight(ds00_device_handle *devh, int value)
{
  return ds00_send(devh,LCD_SET_BRIGHTNESS,value);
}

int ds00_get_fw_version(ds00_device_handle *devh)
{
  return ds00_recv(devh,LCD_GET_FWVER,0);
}

int ds00_set_config(ds00_device_handle *devh, int value)
{
  return ds00_send(devh,LCD_SET_CFG,value);
}

int ds00_set_gpio(ds00_device_handle *devh, int value)
{
  return ds00_send(devh,LCD_SET_GPIO,value);
}

int ds00_cfg_save(ds00_device_handle *devh)
{
  return ds00_send(devh,LCD_SAVE_CFG,0);
}

int ds00_reset(ds00_device_handle *devh, int value)
{
  return ds00_send(devh,LCD_RESET,value);
}

int ds00_get_caps(ds00_device_handle *devh, int idx)
{
  return ds00_recv(devh,LCD_GET_PCB_FEATURES,idx);
}

int ds00_get_config(ds00_device_handle *devh)
{
  return ds00_recv(devh,LCD_GET_CFG,0);
}

int ds00_get_keys(ds00_device_handle *devh)
{
  return ds00_recv(devh,LCD_GET_KEYS,0);
}
